from enum import Enum
from furl import furl

from ppmutils.ppm import PPM
from ppmutils.fhir import FHIR
from ppmutils.settings import ppm_settings

# Get the app logger
import logging

logger = logging.getLogger(ppm_settings.LOGGER_NAME)


class P2MD(PPM.Service):

    service = "data"
    ppm_settings_url_name = "DATA_URL"

    # This is the system prefix used for coding DocumentReferences created by P2MD
    system = FHIR.data_document_reference_identifier_system

    # Set identifier systems
    p2md_identifier_system = "https://peoplepoweredmedicine.org/fhir/p2md/operation"
    fileservice_identifier_system = "https://peoplepoweredmedicine.org/fhir/fileservice/file"

    class ExportProviders(Enum):
        Participant = "ppm-participant"
        i2b2 = "ppm-i2b2"
        JSON = "ppm-json"
        FHIR = "ppm-fhir"

    @classmethod
    def get_auth_link(cls, request, provider, ppm_id, return_url):
        """
        Builds and returns the link that should be followed for the user to authorize
        PPM with the provided data source.
        :param request: The original Django request
        :param provider: The desired data provider (eg Fitbit, Facebook, etc)
        :param ppm_id: The PPM ID of the current user
        :param return_url: The URL to send the user post authorization
        :return: The auth URL
        """

        # Build the base auth URL

        # Build the URL
        url = furl(P2MD._build_url(f"/sources/{provider}/auth/{ppm_id}"))

        # Add the return URL
        url.query.params.add("return_url", return_url)

        # Add the operation name
        url.query.params.add("task", "authorize_{}".format(provider))

        return url.url

    @classmethod
    def get_smart_auth_link(cls, request, provider, ppm_id, return_url):
        """
        Builds and returns the link that should be followed for the user to authorize
        PPM with the provided SMART data source.
        :param request: The original Django request
        :param provider: The desired SMART data provider (eg s4s, epicsandbox, etc)
        :param ppm_id: The PPM ID of the current user
        :param return_url: The URL to send the user post authorization
        :return: The auth URL
        """

        # Build the base auth URL

        # Build the URL
        url = furl(P2MD._build_url(f"/sources/smart/{provider}/auth/{ppm_id}"))

        # Add the return URL
        url.query.params.add("return_url", return_url)

        # Add the operation name
        url.query.params.add("task", "authorize_smart_{}".format(provider))

        return url.url

    @classmethod
    def get_study(cls, request, study):
        """
        Make a request to P2MD to get the current study object. Object
        contains general info about the study.

        :param request: The current HTTP request
        :type request: HttpRequest
        :param study: The study to query on
        :type study: str
        :return: The study object
        :rtype: dict
        """
        # Ensure it's a valid study
        if not study or not PPM.Study.get(study) in PPM.Study:
            raise ValueError(f"'{study}' is not a valid PPM study identifier")

        return cls.get(request, "/ppm/api/study/", {"identifier": PPM.Study.get(study).value})

    @classmethod
    def get_study_is_open(cls, request, study):
        """
        Make a request to P2MD to get data on the study and return whether it
        is open or not.

        :param request: The current HTTP request
        :type request: HttpRequest
        :param study: The study to query on
        :type study: str
        :return: Whether the current study is open or not
        :rtype: boolean
        """
        return next(iter(cls.get_study(request, study)))["is_open"]

    @classmethod
    def get_study_administrators(cls, request, study, support=True, approvals=True):
        """
        Returns a list of the study administrators. Optionally filter based on
        administrator responsibilities.

        :param request: The curren HTTP request
        :type request: HttpRequest
        :param study: The current study
        :type study: str
        :param support: Whether the admin handles support, defaults to True
        :type support: bool, optional
        :param approvals: Whether the admin handles approvals, defaults to True
        :type approvals: bool, optional
        :return: A list of matching administrators
        :rtype: list
        """
        return cls.get(
            request,
            "/ppm/api/administrator/",
            {
                "studies": PPM.Study.get(study).value,
                "does_support": support,
                "does_approval": approvals,
            },
        )

    @classmethod
    def get_authorizations(cls, request, ppm_ids):
        """
        Make a request to P2MD to determine what providers all participants
        have authorized.
        """
        return cls.get(request, "/sources/api/auths", {"ppm_ids": ",".join(ppm_ids)})

    @classmethod
    def has_fitbit_authorization(cls, request, ppm_id):
        """
        Make a request to P2MD to determine if a participant has authorized Fitbit.
        """
        # Return True if no errors
        response = cls.head(request, f"/sources/api/fitbit/{ppm_id}", raw=True)

        return response.ok

    @classmethod
    def has_facebook_authorization(cls, request, ppm_id):
        """
        Make a request to P2MD to determine if a participant has authorized Facebook.
        """
        # Return True if no errors
        response = cls.head(request, f"/sources/api/facebook/{ppm_id}", raw=True)

        return response.ok

    @classmethod
    def has_smart_authorization(cls, request, ppm_id):
        """
        Make a request to P2MD to determine if a participant has authorized
        any SMART provider.
        """
        # Return True if no errors
        response = cls.head(request, f"/sources/api/smart/-/{ppm_id}", raw=True)

        return response.ok

    @classmethod
    def get_smart_authorizations(cls, request, ppm_id):
        """
        Make a request to P2MD to get a list of SMART providers authorized
        by the participant.
        """
        # Make the request
        data = P2MD.get_authorizations(request, [ppm_id])

        # Get list of providers
        auths = next(p["authorizations"] for p in data if p["ppm_id"] == ppm_id)

        # Get list of SMART providers and filter the user's auths list
        smart_providers = [p["provider"] for p in P2MD.get_smart_endpoints(request)["smart_endpoints"]]

        return [auth for auth in auths if auth in smart_providers]

    @classmethod
    def get_operations(cls, request, ppm_id):
        """
        Make a request to P2MD to get a full history of all data operations conducted
        for the participant.
        """
        return cls.get(request, f"/sources/api/ppm/{ppm_id}")

    @classmethod
    def get_twitter_data(cls, request, ppm_id, handle):
        """
        Make a request to P2MD to fetch Twitter data and store it in PPM.
        """
        response = cls.post(request, f"/sources/api/twitter/{ppm_id}", {"handle": handle}, raw=True)

        # Return True if no errors
        return response.ok

    @classmethod
    def download_twitter_data(cls, request, ppm_id):
        """
        Downloads the Twitter dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :return: The requested dataset
        """
        # Make the request
        response = cls.get(request, f"/sources/api/twitter/{ppm_id}/download", raw=True)
        if response:
            return response.content

        return None

    @classmethod
    def get_fitbit_data(cls, request, ppm_id):
        """
        Make a request to P2MD to fetch Fitbit data and store it in PPM.
        """
        response = cls.post(request, f"/sources/api/fitbit/{ppm_id}", data={}, raw=True)

        # Return True if no errors
        return response.ok

    @classmethod
    def download_fitbit_data(cls, request, ppm_id):
        """
        Downloads the Fitbit dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :return: The requested dataset
        """
        # Make the request
        response = cls.get(request, f"/sources/api/fitbit/{ppm_id}/download", raw=True)
        if response:
            return response.content

        return None

    @classmethod
    def get_gencove_data(cls, request, ppm_id, gencove_id):
        """
        Make a request to P2MD to fetch Gencove data and store it in PPM.
        """
        response = cls.post(
            request,
            f"/sources/api/gencove/{ppm_id}",
            data={"gencove_id": gencove_id},
            raw=True,
        )

        # Return True if no errors
        return response.ok

    @classmethod
    def download_gencove_data(cls, request, ppm_id):
        """
        Downloads the Gencove dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :return: The requested dataset
        """
        # Make the request
        response = cls.get(request, f"/sources/api/gencove/{ppm_id}/download", raw=True)
        if response:
            return response.content

        return None

    @classmethod
    def get_facebook_data(cls, request, ppm_id):
        """
        Make a request to P2MD to fetch Facebook data and store it in PPM.
        """
        response = cls.post(request, f"/sources/api/facebook/{ppm_id}", data={}, raw=True)

        # Return True if no errors
        return response.ok

    @classmethod
    def download_facebook_data(cls, request, ppm_id):
        """
        Downloads the Facebook dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :return: The requested dataset
        """
        # Make the request
        response = cls.get(request, f"/sources/api/facebook/{ppm_id}/download", raw=True)
        if response:
            return response.content

        return None

    @classmethod
    def get_smart_data(cls, request, ppm_id, provider):
        """
        Make a request to P2MD to fetch SMART on FHIR EHR data and store it in PPM.
        """
        response = cls.post(request, f"/sources/api/smart/{provider}/{ppm_id}", data={}, raw=True)

        # Return True if no errors
        return response.ok

    @classmethod
    def download_smart_data(cls, request, ppm_id, provider):
        """
        Downloads the SMART dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :param provider: The SMART provider
        :return: The requested entire dataset
        """
        # Make the request
        response = cls.get(request, f"/sources/api/smart/{provider}/{ppm_id}/download", raw=True)
        if response:
            return response.content

        return None

    @classmethod
    def get_files(cls, request, ppm_id):
        """
        Queries P2MD for all uploaded files related to this participant.
        """
        return cls.get(request, f"/sources/api/file/{ppm_id}")

    @classmethod
    def create_file(
        cls,
        request,
        study,
        ppm_id,
        document_type,
        filename,
        metadata=None,
        tags=None,
        content_type="application/octect-stream",
    ):
        """
        Make a request to P2MD to create a file upload
        """
        # Set data
        data = {"study": study, "type": document_type, "filename": filename}

        # Add metadata and tags if passed
        if metadata:
            data["metadata"] = metadata

        if tags:
            data["tags"] = tags

        if content_type:
            data["content_type"] = content_type

        # Get the file data
        upload = cls.post(request, f"/sources/api/file/{ppm_id}", data)

        # Get the UUID
        uuid = upload.get("uuid")

        # Return True if no errors
        return uuid, upload

    @classmethod
    def create_consent_file(cls, request, study, ppm_id, hash, size):
        """
        Make a request to P2MD to create a file upload for a consent PDF
        """
        # Set data
        data = {"study": study, "hash": hash, "size": size}

        # Get the file data
        upload = cls.post(request, f"/sources/api/consent/{study}/{ppm_id}/", data)

        # Get the UUID
        uuid = upload.get("uuid")

        # Return True if no errors
        return uuid, upload

    @classmethod
    def uploaded_consent(cls, request, study, ppm_id, uuid, location):
        """
        Make a request to P2MD to create a file upload
        """
        # Set data
        data = {"study": study, "uuid": uuid, "location": location}

        # Return True if no errors
        return cls.patch(request, f"/sources/api/consent/{study}/{ppm_id}", data)

    @classmethod
    def get_consent_url(cls, study, ppm_id):
        """
        Make a request to P2MD to create a file upload
        """
        # Return True if no errors
        url = cls._build_url(path=f"/sources/api/consent/{study}/{ppm_id}/")

        return url

    @classmethod
    def delete_consent(cls, request, study, ppm_id, document_reference_id):
        """
        Make a request to P2MD to delete a consent render
        """
        # Set data
        data = {"study": study, "document_reference_id": document_reference_id}

        # Return True if no errors
        return cls.delete(request, f"/sources/api/consent/{study}/{ppm_id}", data)

    @classmethod
    def get_qualtrics_surveys(cls, request, study):
        """
        Make a request to P2MD to get available Qualtrics surveys
        :param request: The current request
        :param study: The study for which the surveys should be returned
        :return: list
        """
        # Return response
        return cls.get(request, f"/sources/api/qualtrics/surveys/{study}/")

    @classmethod
    def check_qualtrics_survey(cls, request, study, ppm_id, survey_id):
        """
        Checks the passed survey to see if it has been completed by the
        passed participant or not.
        :param request: The current request
        :param study: The study for which the survey is being used
        :param ppm_id: The participant
        :param survey_id: The ID of the survey to check
        :return: bool
        """
        # Make the request
        response = cls.head(
            request,
            f"/sources/api/qualtrics/survey/{study}/{ppm_id}/{survey_id}/",
            raw=True,
        )
        if response:
            return response.ok

        return False

    @classmethod
    def get_qualtrics_survey_url(cls, study, ppm_id, survey_id):
        """
        Return the URL to send the participant to for taking the survey
        """
        # Return True if no errors
        url = cls._build_url(path=f"/sources/api/qualtrics/survey/{study}/{ppm_id}/{survey_id}/")

        return url

    @classmethod
    def get_qualtrics_survey_data_url(cls, study, ppm_id, survey_id):
        """
        Return the URL to manage survey data
        """
        # Return True if no errors
        url = cls._build_url(path=f"/sources/api/qualtrics/{study}/{ppm_id}/{survey_id}/")

        return url

    @classmethod
    def get_qualtrics_survey_data(cls, request, study, ppm_id, survey_id, response_id=None, older_than=None):
        """
        Make a call to P2MD to look for a survey response
        """
        # Return True if no errors
        url = furl(cls.get_qualtrics_survey_data_url(study=study, ppm_id=ppm_id, survey_id=survey_id))

        data = {}
        if response_id:
            data["response_id"] = response_id

        if older_than:
            data["older_than"] = older_than

        # Make the request
        response = cls.post(request, url.pathstr, data, raw=True)
        if response:
            return response.ok

        return False

    @classmethod
    def get_procure_url(cls, study, ppm_id):
        """
        Return the URL to send users to for Procure data intake
        """
        # Return True if no errors
        url = cls._build_url(path=f"/sources/procure/{study}/{ppm_id}/")

        return url

    @classmethod
    def check_procure(cls, request, study, ppm_id):
        """
        Checks to see if the current participant has already retrieved EHR data via Procure or not
        :param request: The current request
        :param study: The study for which the survey is being used
        :param ppm_id: The participant
        :return: bool
        """
        # Make the request
        response = cls.head(
            request,
            f"/sources/api/procure/{study}/{ppm_id}/",
            raw=True,
        )
        if response:
            return response.ok

        return False

    @classmethod
    def get_procure_manifest_url(cls, study, ppm_id):
        """
        Return the URL to query for details on how to handle fetched FHIR data via Procure
        """
        # Return True if no errors
        url = cls._build_url(path=f"/sources/api/procure/manifest/{study}/{ppm_id}/")

        return url

    @classmethod
    def get_file_proxy_url(cls, ppm_id, uuid):
        """
        Queries P2MD for the download URL for the given file.
        """
        url = cls._build_url(path=f"/sources/api/file/{ppm_id}/{uuid}/")

        return url

    @classmethod
    def uploaded_file(
        cls,
        request,
        study,
        ppm_id,
        document_type,
        uuid,
        location,
        content_type="application/octect-stream",
    ):
        """
        Make a request to P2MD to create a file upload
        """
        # Set data
        data = {
            "study": study,
            "uuid": uuid,
            "location": location,
            "type": document_type,
            "content_type": content_type,
        }

        # Return True if no errors
        return cls.patch(request, f"/sources/api/file/{ppm_id}", data)

    @classmethod
    def get_smart_endpoints(cls, request=None):
        """
        Return a list of all registered SMART endpoints
        :param request: The current HttpRequest
        :return: list
        """
        return cls.get(request, "/sources/api/smart/provider")

    @classmethod
    def get_smart_endpoint_urls(cls, request, ppm_id, return_url):
        """
        Return a list of all registered SMART endpoints with auth links
        for the current user
        :param request: The current HttpRequest
        :param ppm_id: The current user's PPM ID
        :param return_url: The URL to return to post authentication
        :return: list
        """
        smart_endpoints = P2MD.get_smart_endpoints(request)

        # Collect formatted endpoints
        endpoints = []

        # Endpoints
        for endpoint in smart_endpoints.get("smart_endpoints", []):

            # Get required properties
            organization = endpoint.get("organization")
            provider = endpoint.get("provider")
            if not organization or not provider:
                logger.error("Missing properties for SMART endpoint: {} - {}".format(organization, provider))
                continue

            # Build the URL
            url = P2MD.get_smart_auth_link(request, provider, ppm_id, return_url)

            # Add it
            endpoints.append({"organization": organization, "url": url})

        return endpoints

    @classmethod
    def get_providers(cls, request=None):
        """
        Return a list of all registered data providers' codes and titles
        :param request: The current HttpRequest
        :return: list
        """
        return cls.get(request, f"/sources/api/provider")

    @classmethod
    def get_file_types(cls, request=None):
        """
        Return a list of all registered data providers' codes and titles
        :param request: The current HttpRequest
        :return: list
        """
        return cls.get(request, f"/sources/api/file/type")

    @classmethod
    def get_participant_data_url(cls, ppm_id, filename=None, providers=None):
        """
        Downloads the PPM dataset for the passed user
        :param ppm_id: The PPM ID of the requesting user
        :param filename: What the resulting archive should be called
        :param providers: A list of providers to limit data included to
        :return: The user's entire dataset
        """
        # Build the URL
        url = furl(P2MD._build_url(f"/sources/api/archive/{ppm_id}/download/"))

        # If providers, include them in query
        if providers:
            url.query.params.add("providers", ",".join(providers))

        # Add the return URL
        if filename:
            url.query.params.add("filename", filename)

        return url.url

    @classmethod
    def download_participant_data(cls, request, ppm_id, filename=None, providers=None):
        """
        Downloads the PPM dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :param filename: What the resulting archive should be called
        :param providers: A list of providers to limit data included to
        :return: The user's entire dataset
        """
        # Build the URL
        url = furl(P2MD.get_participant_data_url(ppm_id=ppm_id, filename=filename, providers=providers))

        # Make the request
        response = cls.get(request=request, path=url.pathstr, data=url.querystr, raw=True)
        if response:
            return response.content

        return None

    @classmethod
    def export_content_type(cls, provider=ExportProviders.Participant):
        """
        Returns the content type and extension for the passed export provider.
        Use this method to prepare a response when passing through a
        Fileservice document.
        :param provider: ExportProvider
        :return: str, str
        """
        if provider is cls.ExportProviders.Participant:
            return "application/zip", "zip"

        else:
            return "application/json", "json"

    @classmethod
    def get_data_document_references(cls, ppm_id, provider=None):
        """
        Queries the current user's FHIR record for any DocumentReferences
        related to this type
        :return: A list of DocumentReferences
        :rtype: list
        """
        # Gather data-related DocumentReferences
        document_references = FHIR.query_data_document_references(patient=ppm_id, provider=provider)
        logger.debug(f"{ppm_id}: Found {len(document_references)} DocumentReferences " f"for: {provider}")

        # Flatten resources and pick out relevant identifiers
        flats = []
        for document_reference in document_references:

            # Flatten it
            flat = FHIR.flatten_document_reference(document_reference)

            # Pick out P2MD and Fileservice identifiers
            if P2MD.p2md_identifier_system in flat:
                flat["p2md_id"] = flat[P2MD.p2md_identifier_system]
                del flat[P2MD.p2md_identifier_system]

            if P2MD.fileservice_identifier_system in flat:
                flat["fileservice_id"] = flat[P2MD.fileservice_identifier_system]
                del flat[P2MD.fileservice_identifier_system]

            elif flat.get("url"):
                # To support older documents, try to parse it from URL
                url = furl(flat["url"])
                flat["fileservice_id"] = url.path.segments.pop(3)

            else:
                # Just make it empty
                flat["fileservice_id"] = "ERROR"

            flats.append(flat)

        return flats

    @classmethod
    def get_data_document_references_for_providers(cls, ppm_id, providers=None):
        """
        Queries the current user's FHIR record for any DocumentReferences related
        to the passed types
        :return: A list of DocumentReferences
        :rtype: list
        """
        # Get all flattened data document references
        document_references = []
        for document_reference in P2MD.get_data_document_references(ppm_id, provider=None):

            # Check type and filter out non-requested provider documents
            if not providers or document_reference["type"] not in providers:
                continue

            document_references.append(document_reference)

        logger.debug(f"{ppm_id}: Found {len(document_references)} DocumentReferences " f'for: {", ".join(providers)}')
        return document_references

    #
    # Deprecated
    #

    @classmethod
    def check_export(cls, request, ppm_id, provider=ExportProviders.Participant, age=24):
        """
        Checks the presence of the PPM dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :param provider: The provider or format of the exported data
        :param age: Set the number of hours after which the dataset should be
        considered expired
        :return: The age of the current dataset in hours, if any
        """
        # Make the request
        response = cls.head(
            request,
            f"/sources/api/ppm/{provider.value}/{ppm_id}/export",
            {"age": age},
            raw=True,
        )
        if response:
            return response.ok

        return False

    @classmethod
    def get_export_url(cls, request, ppm_id, provider=ExportProviders.Participant):
        """
        Generates the URL to download the PPM dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :param provider: The provider or format of the exported data
        :return: The user's entire dataset
        """
        url = furl(cls._build_url(f"/sources/api/ppm/{provider.value}/{ppm_id}/export"))

        return url.url

    @classmethod
    def download_export(cls, request, ppm_id, provider=ExportProviders.Participant):
        """
        Downloads the PPM dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :param provider: The provider or format of the exported data
        :return: The user's entire dataset
        """
        # Make the request
        response = cls.get(request, f"/sources/api/ppm/{provider.value}/{ppm_id}/export", raw=True)
        if response:
            return response.content

        return None

    @classmethod
    def download_data(cls, request, ppm_id, provider=ExportProviders.Participant):
        """
        Downloads the PPM dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :param provider: The provider or format of the exported data
        :return: The user's entire dataset
        """
        # Make the request
        response = cls.get(request, f"/sources/api/ppm/{provider.value}/{ppm_id}/download", raw=True)
        if response:
            return response.content

        return None

    @classmethod
    def download_data_notify(cls, request, ppm_id, recipients, provider=ExportProviders.Participant):
        """
        Downloads the PPM dataset for the passed user
        :param request: The original Django request object
        :param ppm_id: The PPM ID of the requesting user
        :param recipients: The email addresses to which a notification of the
        data's preparation should be sent
        :param provider: The provider or format of the exported data
        :return: The user's entire dataset
        """
        # Make the request
        response = cls.post(
            request,
            path=f"/sources/api/ppm/{provider.value}/{ppm_id}/",
            data={"recipients": ",".join(recipients)},
            raw=True,
        )

        return response.ok



from ppmutils.ppm import PPM
from ppmutils.settings import ppm_settings

# Get the app logger
import logging

logger = logging.getLogger(ppm_settings.LOGGER_NAME)


class API(PPM.Service):

    service = "ppm"
    proxied = True

    @classmethod
    def studies(cls, as_choices=False):
        """
        Queries the API for the dashboard for the given study
        :param study: The study code
        :param as_choices: Return the items as a tuple of tuples
        :return: dict
        """
        # Make the request
        items = cls.get(request=None, path=f"/study/")

        # Return dashboard
        return items if not as_choices else ((i["code"], i["title"]) for i in items)

    @classmethod
    def dashboard(cls, study=None):
        """
        Queries the API for the dashboard for the given study
        :param study: The study code
        :return: dict
        """
        # Make the request
        if not study:
            items = cls.get(request=None, path=f"/step/")
        else:
            # Get the study proper
            study = PPM.Study.get(study)

            items = cls.get(request=None, path=f"/study/{study.value}/")["dashboard"]

        return items

    @classmethod
    def enrollments(cls, as_choices=False):
        """
        Queries the API for the dashboard for the given study
        :param as_choices: Return the items as a tuple of tuples
        :return: dict
        """
        # Make the request
        items = cls.get(request=None, path=f"/enrollment/")

        # Return dashboard
        return items if not as_choices else ((i["code"], i["title"]) for i in items)

    @classmethod
    def samples(cls, study=None, as_choices=False):
        """
        Queries the API for PPM samples
        :param study: Filter samples for the given study
        :param as_choices: Return the items as a tuple of tuples
        :return: dict
        """
        # Make the request
        if not study:
            items = cls.get(request=None, path=f"/sample/")
        else:
            # Get the study proper
            study = PPM.Study.get(study)

            items = cls.get(request=None, path=f"/study/{study.value}/")["samples"]

        # Return dashboard
        return items if not as_choices else ((i["code"], i["title"]) for i in items)



#!/usr/bin/env python
"""
This script is a trick to setup a fake Django environment, since this reusable
app will be developed and tested outside any specific Django project.
Via ``settings.configure`` you will be able to set all necessary settings
for your app and run the tests as if you were calling ``./manage.py test``.
"""
import sys
from django_nose import NoseTestSuiteRunner
from django.conf import settings

EXTERNAL_APPS = [
    "django.contrib.admin",
    "django.contrib.admindocs",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.sitemaps",
    "django.contrib.sites",
]
INTERNAL_APPS = [
    "django_nose",
    "ppmutils",
]
INSTALLED_APPS = EXTERNAL_APPS + INTERNAL_APPS

if not settings.configured:
    settings.configure(
        DATABASES={"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:",}},
        INSTALLED_APPS=INSTALLED_APPS,
        PPM_CONFIG={"FHIR_URL": "https://fhir.ppm.dbmi.hms.harvard.edu/baseDstu3"},
    )


def main(*test_args):
    sys.exit(NoseTestSuiteRunner(verbosity=2, interactive=True).run_tests(test_args))


if __name__ == "__main__":
    main(*sys.argv[1:])



import os
import csv
import json
import io
from enum import Enum

import logging

logger = logging.getLogger(__name__)


class PicnicHealth(object):
    """ Facilitates the processing and transformation of a PicnicHealth CSV export into a PPM usable format """

    # Define PicnicHealth resource types
    class Resource(Enum):
        CareSite = "careSite"
        Concept = "concept"
        Condition = "condition"
        ConditionReadable = "conditionReadable"
        DicomStudy = "dicomStudy"
        Drug = "drug"
        DrugReadable = "drugReadable"
        Location = "location"
        Measurement = "measurement"
        MeasurementReadable = "measurementReadable"
        PdfFile = "pdfFile"
        Person = "person"
        Provider = "provider"
        Visit = "visit"
        Note = "note"

        @classmethod
        def person_resources(cls):
            """ Return a list of resource types that link to Person objects directly """
            return [
                r
                for r in PicnicHealth.Resource
                if r not in PicnicHealth.Resource.visit_resources() + PicnicHealth.Resource.care_site_resources()
            ]

        @classmethod
        def visit_resources(cls):
            """ Return a list of resource types that link to Visit objects """
            return [PicnicHealth.Resource.CareSite, PicnicHealth.Resource.Provider]

        @classmethod
        def care_site_resources(cls):
            """ Return a list of resource types that link to Care Site objects """
            return [
                PicnicHealth.Resource.Location,
            ]

    def __init__(self, directory, output, force=False, dry=False):
        self.directory = directory
        self.output = output
        self.force = force
        self.dry = dry

    def _path_for_resource(self, resource):
        """
        Returns the path to the resource type's CSV file, considering the current working directory
        :param resource: A PicnicHealth resource type
        :type resource: PicnicHealth.Resource
        :return: The path to the CSV
        :rtype: str
        """
        return os.path.join(self.directory, "{}.csv".format(resource.value))

    def prepare_output_path(self, path):
        """
        Accepts a path and prepares the directory for files to be written from the conversion process
        :param path: The path to check
        :return: Whether the output path is ready or not
        """
        # Check if exists
        if os.path.exists(path) and input("File exists at '{}', overwrite contents (n)?".format(path)) not in [
            "y",
            "yes",
        ]:
            return False
        elif not os.path.exists(os.path.dirname(path)):
            try:
                if not self.dry:
                    os.mkdir(os.path.dirname(path))
            except Exception:
                raise ValueError("Could not create owner directory '{}'".format(os.path.dirname(path)))

        return True

    def get_sheet(self, resource):
        """
        Accepts a PicnicHealth resource type and attempts to read in its sheet
        :param resource: A PicnicHealth resource type
        :type resource: PicnicHealth.Resource
        :return: The sheet, if read successfully
        :rtype: csv.DictReader
        """
        try:
            # Check path
            path = self._path_for_resource(resource)
            if not os.path.exists(path):
                print(f'Error: File "{path}" for {resource.name} does not exist')
                return None

            # Dump the JSON document
            with open(path, "r") as f:
                yield csv.DictReader(f)

        except Exception as e:
            print(f"Error: {e}")

        return None

    def extract_resources(self, resource, key=None, value=None, values=None):
        """
        Accepts a csv.DictReader and extracts the rows that match the passed key (fieldname) and value. Returns a list
        of dictionary objects parsed from the sheet's matching rows.
        :param resource: A PicnicHealth resource type whose CSV should be looked into for matching rows
        :type resource: PicnicHealth.Resource
        :param key: The key or fieldname to perform matching lookups on
        :type key: str
        :param value: The value that the passed key or fieldname should match in order to confirm a
        matching row/resource
        :type value: str
        :param values: The values that the passed key or fieldname should match in order to confirm a
        matching row/resource
        :type values: list
        :return: A list of objects
        :rtype: list
        """
        f = None
        try:
            # Check value/values
            if value is not None and values is not None:
                print('Warning: Both "value" and "values" are specified, priority is given to "value"')

            # Check path
            path = self._path_for_resource(resource)
            if not os.path.exists(path):
                print(f'Error: File "{path}" for {resource.name} does not exist')
                return None

            # Dump the JSON document
            f = open(path, "r")
            sheet = csv.DictReader(f)

            # Collect resources
            resources = []

            # Find those related to this person
            for sheet_row in sheet:
                # Compare to value, if passed
                if key is not None:
                    if value is not None and key in sheet_row and sheet_row[key] != value:
                        continue
                    # Compare to values, if passed
                    elif values is not None and key in sheet_row and sheet_row[key] not in values:
                        continue

                # Add it to JSON
                resource = {}
                for fieldname in iter(sheet.fieldnames):
                    try:
                        # If it's a nested JSON object, parse it first
                        resource[fieldname] = json.loads(sheet_row[fieldname])
                    except json.JSONDecodeError:
                        # Else, just copy it over
                        resource[fieldname] = sheet_row[fieldname]

                # Add it
                resources.append(resource)

            # Close file
            f.close()

            return resources

        except Exception as e:
            logger.exception(f"Error: {e}", exc_info=True)
            # Clean up
            if f:
                f.close()

        return None

    def output_csv(self, resource, owner, resources):
        """
        Accepts a list of resources and exports the list of resources to a CSV file at the passed path
        :param resource: A PicnicHealth resource type whose CSV should be looked into for matching rows
        :type resource: PicnicHealth.Resource
        :param owner: The owner's ID, determines where files are saved in the output
        :type owner: str
        :param resources: A list of dictionary objects describing rows for the outgoing CSV
        :type resources: list
        :return: Whether the write performed successfully or not
        :rtype: bool
        """
        output_f = None
        try:
            # Determine output path
            path = os.path.join(self.output, owner, "{}.csv".format(resource.value))

            # Prepare output path
            if not self.prepare_output_path(path=path):
                return False

            # Collect fieldnames
            fieldnames = []
            for resource in resources:
                fieldnames = list(set(fieldnames + list(resource.keys())))

            # Start the writer
            output_f = open(path, "w") if not self.dry else io.StringIO()
            sheet_writer = csv.DictWriter(output_f, fieldnames=fieldnames)
            sheet_writer.writeheader()

            # Find those related to this person
            for resource in resources:

                # Write it
                sheet_writer.writerow(resource)

            # Check dry run
            if self.dry:
                print('DRY: Would have saved resources to "{}"'.format(path))

            # Close files
            output_f.close()

            return True

        except Exception as e:
            logger.exception(f"Error: {e}", exc_info=True)
            # Clean up
            if output_f:
                output_f.close()

        return False

    def split_csv(self, resource, owner, key=None, value=None, values=None):
        """
        Accepts a csv.DictReader and extracts the rows that match the passed key (fieldname) and value. Matching
        rows are copied to a new CSV at the passed path.
        :param resource: A PicnicHealth resource type whose CSV should be looked into for matching rows
        :type resource: PicnicHealth.Resource
        :param owner: The owner's ID, determines where files are saved in the output
        :type owner: str
        :param key: The key or fieldname to perform matching lookups on
        :type key: str
        :param value: The value that the passed key or fieldname should match in order to confirm a
        matching row/resource
        :type value: str
        :param values: The values that the passed key or fieldname should match in order to confirm a
        matching row/resource
        :type values: list
        :return: Whether the write performed successfully or not
        :rtype: bool
        """
        f = None
        output_f = None
        try:
            # Determine output path
            path = os.path.join(self.output, owner, "{}.csv".format(resource.value))

            # Prepare output path
            if not self.prepare_output_path(path=path):
                return False

            # Check path
            sheet_path = self._path_for_resource(resource)
            if not os.path.exists(sheet_path):
                print(f'Error: File "{sheet_path}" for {resource.name} does not exist')
                return None

            # Dump the JSON document
            f = open(sheet_path, "r")
            sheet = csv.DictReader(f)

            # Start the writer
            output_f = open(path, "w") if not self.dry else io.StringIO()
            sheet_writer = csv.DictWriter(output_f, fieldnames=sheet.fieldnames)
            sheet_writer.writeheader()

            # Find those related to this person
            rows = 0
            for sheet_row in sheet:
                # Compare to value, if passed
                if key is not None:
                    if value is not None and key in sheet_row and sheet_row[key] != value:
                        continue
                    # Compare to values, if passed
                    elif values is not None and key in sheet_row and sheet_row[key] not in values:
                        continue

                # Write it
                sheet_writer.writerow(sheet_row)
                rows += 1

            # Check dry run
            if self.dry:
                print('DRY: Would have split {} rows for "{}"="{}" -> "{}"'.format(rows, key, value, path))

            # Close files
            output_f.close()

            return True

        except Exception as e:
            logger.exception(f"Error: {e}", exc_info=True)
            # Clean up
            if output_f:
                output_f.close()
            if f:
                f.close()

        return False

    def output_json(self, resource, owner, resources):
        """
        Accepts a list of resources and exports the list of resources to a JSON file at the passed path
        :param resource: A PicnicHealth resource type whose CSV should be looked into for matching rows
        :type resource: PicnicHealth.Resource
        :param owner: The owner's ID, determines where files are saved in the output
        :type owner: str
        :param resources: A list of dictionary objects describing rows for the outgoing JSON
        :type resources: list
        :return: Whether the write performed successfully or not
        :rtype: bool
        """
        f = None
        try:
            # Get the path
            path = os.path.join(self.output, owner, "{}.json".format(resource.value))

            # Prepare output path
            if not self.prepare_output_path(path=path):
                return False

            # Get the file handle to write to
            f = open(path, "w") if not self.dry else io.StringIO()

            # Dump the JSON document
            json.dump(resources, f)

            # Check dryrun
            if self.dry:
                print('DRY: Would have written resources to "{}"'.format(path))

            # Close it
            f.close()

            return True

        except Exception as e:
            logger.exception(f"Error: {e}", exc_info=True)
            # Clean up
            if f:
                f.close()

        return False

    def csv(self):
        """
        Runs the conversion operation on the input directory's CSV sheets. Outputs only CSV
        each person in their own directory within the specified output directory
        """
        # Get Persons
        persons = self.extract_resources(resource=PicnicHealth.Resource.Person)
        for person in persons:

            # Get their ID create a placeholder for visits
            person_id = person["personId"]
            provider_ids = None
            care_site_ids = None

            # Extract resources
            for resource in PicnicHealth.Resource.person_resources():

                # Split
                self.split_csv(resource=resource, owner=person_id, key="personId", value=person_id)

                # If visits, pull linked resource IDs
                if resource is PicnicHealth.Resource.Visit:

                    # Get resources and extract provider and care site IDs
                    resources = self.extract_resources(resource=resource, key="personId", value=person_id)
                    provider_ids = list(
                        set(
                            [r.get("performingProviderId") for r in resources if r.get("performingProviderId")]
                            + [r.get("referringProviderId") for r in resources if r.get("referringProviderId")]
                        )
                    )
                    care_site_ids = list(set([r.get("careSiteId") for r in resources if r.get("careSiteId")]))

            # Get location IDs now since we needed care sites to be processed
            resources = self.extract_resources(
                resource=PicnicHealth.Resource.CareSite, key="careSiteId", values=care_site_ids
            )
            location_ids = list(set([r.get("locationId") for r in resources if r.get("locationId")]))

            # Split the rest
            self.split_csv(
                resource=PicnicHealth.Resource.Provider, owner=person_id, key="providerId", values=provider_ids
            )
            self.split_csv(
                resource=PicnicHealth.Resource.CareSite, owner=person_id, key="careSiteId", values=care_site_ids
            )
            self.split_csv(
                resource=PicnicHealth.Resource.Location, owner=person_id, key="locationId", values=location_ids
            )

    def json(self, write=True):
        """
        Runs the conversion operation on the input directory's CSV sheets. Outputs JSON for
        each person in their own directory within the specified output directory
        :param write: If set, files will be written to the configured output directory. If not, StringIO objects are
        returned
        :return: True if write and write succeeds, StringIO objects if not writing
        """
        # Get Persons
        persons = self.extract_resources(resource=PicnicHealth.Resource.Person)
        for person in persons:

            # Get their ID create a placeholder for visits
            person_id = person["personId"]
            provider_ids = None
            care_site_ids = None
            location_ids = None

            # Extract resources
            for resource in PicnicHealth.Resource.person_resources():

                # Extract
                resources = self.extract_resources(resource=resource, key="personId", value=person_id)
                if not resources:
                    continue

                # Export to JSON
                self.output_json(resource=resource, owner=person_id, resources=resources)

                # Save them if visits
                if resource is PicnicHealth.Resource.Visit:

                    # Get provider and care site IDs
                    provider_ids = list(
                        set(
                            [r.get("performingProviderId") for r in resources if r.get("performingProviderId")]
                            + [r.get("referringProviderId") for r in resources if r.get("referringProviderId")]
                        )
                    )
                    care_site_ids = list(set([r.get("careSiteId") for r in resources if r.get("careSiteId")]))

                elif resource is PicnicHealth.Resource.CareSite:

                    # Get location IDs
                    location_ids = list(set([r.get("locationId") for r in resources if r.get("locationId")]))

            # Process providers
            providers = self.extract_resources(
                resource=PicnicHealth.Resource.Provider, key="providerId", values=provider_ids
            )

            # Export to JSON
            self.output_json(resource=PicnicHealth.Resource.Provider, owner=person_id, resources=providers)

            # Process care sites
            care_sites = self.extract_resources(
                resource=PicnicHealth.Resource.CareSite, key="careSiteId", values=care_site_ids
            )

            # Export to JSON
            self.output_json(resource=PicnicHealth.Resource.CareSite, owner=person_id, resources=care_sites)

            # Process locations
            locations = self.extract_resources(
                resource=PicnicHealth.Resource.Location, key="locationId", values=location_ids
            )

            # Export to JSON
            self.output_json(resource=PicnicHealth.Resource.Location, owner=person_id, resources=locations)
